// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
//
// Die Pinbelegung ist ber defines in lcdlib.h einstellbar

// HD44780 LCD compatible library 
// The pin layout can be defined in the lcdlib.h

#include <avr/io.h>
#include "lcdlib.h"
#include <util/delay.h>

////////////////////////////////////////////////////////////////////////////////
// Erzeugt einen Enable-Puls	;generates an enable pulse on the enable-pin
static void lcd_enable( void )
{
	LCD_PORT |= (1<<LCD_EN);     // Enable auf 1 setzen ; Set enable 1
	_delay_us( LCD_ENABLE_US );  // kurze Pause ; short break
	LCD_PORT &= ~(1<<LCD_EN);    // Enable auf 0 setzen ; Set enable 0
}

////////////////////////////////////////////////////////////////////////////////
// Sendet eine 4-bit Ausgabeoperation an das LCD ; Sending a 4-bit Output operation
static void lcd_out( uint8_t data )
{
	data &= 0xF0;                       // obere 4 Bit maskieren ; mask upper 4 bits
	
	LCD_PORT &= ~(0xF0>>(4-LCD_DB));    // Maske lschen ; delete mask
	LCD_PORT |= (data>>(4-LCD_DB));     // Bits setzen ; set bits
	lcd_enable();
}

////////////////////////////////////////////////////////////////////////////////
// Initialisierung: muss ganz am Anfang des Programms aufgerufen werden. ; Initialization: has to be called/invoked at the beginning of the program!
void lcd_init( void )
{
	// verwendete Pins auf Ausgang schalten		;Set required pins to output
	uint8_t pins = (0x0F << LCD_DB) |           // 4 Datenleitungen ; 4 data connections (Pins)
	(1<<LCD_RS) |                // R/S Leitung ; Reset/Set connection (Pin)
	(1<<LCD_EN);                 // Enable Leitung ; Enable connection (Pin)
	LCD_DDR |= pins;
	
	// initial alle Ausgnge auf Null ; Initially, all outputs to 0
	LCD_PORT &= ~pins;
	
	// warten auf die Bereitschaft des LCD ; Waiting for LCD readiness
	_delay_ms( LCD_BOOTUP_MS );
	
	// Soft-Reset muss 3mal hintereinander gesendet werden zur Initialisierung ; Soft-Reset, has to be called three times in a row to initialize. (Data sheet)
	lcd_out( LCD_SOFT_RESET );
	_delay_ms( LCD_SOFT_RESET_MS1 );
	
	lcd_enable();
	_delay_ms( LCD_SOFT_RESET_MS2 );
	
	lcd_enable();
	_delay_ms( LCD_SOFT_RESET_MS3 );
	
	// 4-bit Modus aktivieren ; Activate 4-bit mode
	lcd_out( LCD_SET_FUNCTION |
	LCD_FUNCTION_4BIT );
	_delay_ms( LCD_SET_4BITMODE_MS );
	
	// 4-bit Modus / 2 Zeilen / 5x7 ; 4-bit mode / 2 rows / 5x7
	lcd_command( LCD_SET_FUNCTION |
	LCD_FUNCTION_4BIT |
	LCD_FUNCTION_2LINE |
	LCD_FUNCTION_5X7 );
	
	// Display ein / Cursor aus / Blinken aus ; Display on / cursor off / blinking off
	lcd_command( LCD_SET_DISPLAY |
	LCD_DISPLAY_ON |
	LCD_CURSOR_OFF |
	LCD_BLINKING_OFF);
	
	// Cursor inkrement / kein Scrollen ; Inrement cursor / no scroll
	lcd_command( LCD_SET_ENTRY |
	LCD_ENTRY_INCREASE |
	LCD_ENTRY_NOSHIFT );
	
	lcd_clear();
}

////////////////////////////////////////////////////////////////////////////////
// Sendet ein Datenbyte an das LCD ; Sends a data byte to the LCD
void lcd_data( uint8_t data )
{
	LCD_PORT |= (1<<LCD_RS);    // RS auf 1 setzen ; Set RS to 1
	
	lcd_out( data );            // zuerst die oberen,  ; first send the upper 4 bits
	lcd_out( data<<4 );         // dann die unteren 4 Bit senden ; then the lower 4 bits
	
	_delay_us( LCD_WRITEDATA_US );
}

////////////////////////////////////////////////////////////////////////////////
// Sendet einen Befehl an das LCD ; Sends a command to the LCD
void lcd_command( uint8_t data )
{
	LCD_PORT &= ~(1<<LCD_RS);    // RS auf 0 setzen ; Set RS to 0
	
	lcd_out( data );             // zuerst die oberen, ; first send the upper 4 bits
	lcd_out( data<<4 );           // dann die unteren 4 Bit senden ; then the lower 4 bits
	
	_delay_us( LCD_COMMAND_US );
}

////////////////////////////////////////////////////////////////////////////////
// Sendet den Befehl zur Lschung des Displays 	;Sends a clear display command
void lcd_clear( void )
{
	lcd_command( LCD_CLEAR_DISPLAY );
	_delay_ms( LCD_CLEAR_DISPLAY_MS );
}

////////////////////////////////////////////////////////////////////////////////
// Sendet den Befehl: Cursor Home 	;Sends the command "Cursor home"
void lcd_home( void )
{
	lcd_command( LCD_CURSOR_HOME );
	_delay_ms( LCD_CURSOR_HOME_MS );
}

////////////////////////////////////////////////////////////////////////////////
// Setzt den Cursor in Spalte x (0..15) Zeile y (1..4)	;Sets the cursor into column x (0..15) and row y (1..4)

void lcd_setcursor( uint8_t x, uint8_t y )
{
	uint8_t data;
	
	switch (y)
	{
		case 1:    // 1. Zeile	;1.st row
		data = LCD_SET_DDADR + LCD_DDADR_LINE1 + x;
		break;
		
		case 2:    // 2. Zeile	;2nd row
		data = LCD_SET_DDADR + LCD_DDADR_LINE2 + x;
		break;
		
		case 3:    // 3. Zeile	;3rd row
		data = LCD_SET_DDADR + LCD_DDADR_LINE3 + x;
		break;
		
		case 4:    // 4. Zeile	;4th row
		data = LCD_SET_DDADR + LCD_DDADR_LINE4 + x;
		break;
		
		default:
		return;                                   // fr den Fall einer falschen Zeile 	;In case of a wrong x (row) value
	}
	
	lcd_command( data );
}

////////////////////////////////////////////////////////////////////////////////
// Schreibt einen String auf das LCD	;Writes a string to the display

void lcd_string( const char *data )
{
	while( *data != '\0' )
	lcd_data( *data++ );
}

////////////////////////////////////////////////////////////////////////////////
// Schreibt ein Zeichen in den Character Generator RAM 	;Writes a character into the character generator RAM of the LCD

void lcd_generatechar( uint8_t code, const uint8_t *data )
{
	// Startposition des Zeichens einstellen
	lcd_command( LCD_SET_CGADR | (code<<3) );
	
	// Bitmuster bertragen
	for ( uint8_t i=0; i<8; i++ )
	{
		lcd_data( data[i] );
	}
}